"""The components of the derivative of the entropy for the Choo and Siow homoskedastic model."""

import numpy as np
from typing import Optional, Union, Tuple

from .utils import bs_error_abort
from .matching_utils import _get_singles

from .entropy import EntropyHessianComponents, EntropyFunctions


def _entropy_choo_siow(muxyhat: np.ndarray,
                       n: np.ndarray, m: np.ndarray, deriv: Optional[int] = 0) -> \
        Union[np.ndarray, Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """Returns the values of :math:`\\mathcal{E}` at :math:`(\\hat{\\mu}_{xy})`
    and the first (if `deriv` is 1 or 2) and second (if `deriv` is 2) derivatives
    for the Choo and Siow model

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins
        deriv: if equal 1, we compute the first derivatives too; if equals 2, also the hessian

    Returns:
        the value of the generalized entropy
        if deriv = 1 or 2, the (X,Y) matrix of the first derivative of the entropy
        if deriv = 2, the (X,Y,X,Y) array of the second derivative
    """
    muxy = muxyhat.astype(float)
    mux0, mu0y = _get_singles(muxy, n, m)

    logxy = np.log(muxy)
    logx0 = np.log(mux0)
    log0y = np.log(mu0y)

    val_entropy = -2.0 * np.sum(muxy * logxy) - np.sum(mux0 * logx0) - np.sum(mu0y * log0y) \
                  + np.sum(n * np.log(n)) + np.sum(m * np.log(m))

    if deriv == 0:
        return val_entropy
    if deriv in [1, 2]:
        der_xy = -2.0 * logxy + log0y
        der_xy += logx0.reshape((-1, 1))
        if deriv == 1:
            return val_entropy, der_xy
        else:  # we compute the Hessian
            X, Y = muxyhat.shape
            derlogxy = 1.0 / muxy
            derlogx0 = 1.0 / mux0
            derlog0y = 1.0 / mu0y
            der2_xyzt = np.zeros((X, Y, X, Y))
            for x in range(X):
                dlogx0 = derlogx0[x]
                for y in range(Y):
                    d2xy = np.zeros((X, Y))
                    d2xy[x, :] = - dlogx0
                    d2xy[:, y] -= derlog0y[y]
                    d2xy[x, y] -= 2.0 * derlogxy[x, y]
                    der2_xyzt[x, y, :, :] = d2xy
            return val_entropy, der_xy, der2_xyzt
    else:
        bs_error_abort(f"deriv should be 0, 1, or 2")


def e0_fun_choo_siow(muxyhat, n, m):
    """Returns the values of t :math:`e_0` at :math:`\\hat{\\mu}` for the Choo and Siow model.

        Args:
            muxyhat: an (X, Y) matrix of matching patterns
            n: an X-vector of margins
            m: an Y-vector of margins

        Returns:
            the (X,Y) matrix of the first derivative of the entropy
    """
    entropy_res = _entropy_choo_siow(muxyhat, n, m, deriv=1)
    return entropy_res[1]


def e0_derivative_choo_siow(muxyhat, n, m) -> EntropyHessianComponents:
    """Returns the derivatives of :math:`e_0` at :math:`\\hat{\\mu}` for the Choo and Siow model.

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        the three components of the hessian of the entropy
    """
    entropy_res = _entropy_choo_siow(muxyhat, n, m, deriv=2)
    hess = entropy_res[2]
    X, Y = muxyhat.shape
    hess_x = np.zeros((X, Y, Y))
    hess_y = np.zeros((X, Y, X))
    hess_xy = np.zeros((X, Y))
    for x in range(X):
        for y in range(Y):
            d2xy = hess[x, y, :, :]
            hess_x[x, y, :] = d2xy[x, :]
            hess_y[x, y, :] = d2xy[:, y]
            hess_xy[x, y] = d2xy[x, y]
    return hess_x, hess_y, hess_xy


entropy_choo_siow = EntropyFunctions(e0_fun=e0_fun_choo_siow,
                                     hessian="provided", e0_derivative=e0_derivative_choo_siow,
                                     description="Choo and Siow homoskedastic with analytic Hessian")

entropy_choo_siow_numeric = EntropyFunctions(e0_fun=e0_fun_choo_siow,
                                             description="Choo and Siow homoskedastic with numerical Hessian")
