""" matching-related utilities """

from dataclasses import dataclass
import numpy as np
import scipy.stats as sts
from typing import Tuple, List

from .utils import print_stars, bs_error_abort, test_vector, test_matrix


@dataclass
class Matching:
    """ stores the numbers of couples and singles of every type;

    muxy is an (X,Y)-matrix
    mux0 is an X-vector
    mu0y is an Y-vector
    """
    muxy: np.ndarray
    mux0: np.ndarray
    mu0y: np.ndarray

    def __repr__(self):
        X, Y = self.muxy.shape
        n_couples = np.sum(self.muxy)
        n_single_men, n_single_women = np.sum(self.mux0), np.sum(self.mu0y)
        repr_str = f"This is a matching with {n_couples} couples,\n {n_single_men} single men,\n"
        repr_str += f"   and {n_single_women} single women.\n"
        repr_str += f" We have {X} types of men and {Y} of women."
        print_stars(repr_str)

    def __post_init__(self):
        X, Y = test_matrix(self.muxy)
        Xn = test_vector(self.mux0)
        Ym = test_vector(self.mu0y)
        if Xn != X:
            bs_error_abort(
                f"muxy is a ({X}, {Y}) matrix but mux0 has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(
                f"muxy is a ({X}, {Y}) matrix but mu0y has {Ym} elements.")

    def unpack(self):
        muxy, mux0, mu0y = self.muxy, self.mux0, self.mu0y
        n = np.sum(muxy, 1) + mux0
        m = np.sum(muxy, 0) + mu0y
        return muxy, mux0, mu0y, n, m


def _get_singles(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Computes the numbers of singles from the matches and the margins."""
    mux0hat = n - np.sum(muxyhat, 1)
    mu0yhat = m - np.sum(muxyhat, 0)
    return mux0hat, mu0yhat


def _get_margins(mus: Matching) -> Tuple[np.ndarray, np.ndarray]:
    """ computes the numbers of each type from the matching patterns"""
    _, _, _, n, m = mus.unpack()
    return n, m


def _simulate_sample_from_mus(mus: Matching, n_households: int) -> Tuple[Matching, np.ndarray, np.ndarray]:
    """Draw a sample of n_households form the matching patterns in mus

        Args:
            mus: the matching patterns
            n_households: the number of households requested

        Returns:
            the sample matching patterns and the sample margins
    """
    muxy, mux0, mu0y, _, _ = mus.unpack()
    X, Y = muxy.shape
    # stack all probabilities
    XY = X * Y
    num_choices = XY + X + Y
    pvec = np.zeros(num_choices)
    pvec[:XY] = muxy.reshape(XY)
    pvec[XY:(XY + X)] = mux0
    pvec[(XY + X):] = mu0y
    pvec /= np.sum(pvec)
    matches = sts.multinomial.rvs(n_households, p=pvec)
    muxy_sim = matches[:XY].reshape((X, Y))
    mux0_sim = matches[XY:(XY + X)]
    mu0y_sim = matches[(XY + X):]
    # make sure we have no zeros
    _MU_EPS = min(1, int(1e-3 * n_households))
    muxy_sim += _MU_EPS
    mux0_sim += _MU_EPS
    mu0y_sim += _MU_EPS
    mus_sim = Matching(muxy=muxy_sim, mux0=mux0_sim, mu0y=mu0y_sim)
    n_sim, m_sim = _get_margins(mus_sim)
    return mus_sim, n_sim, m_sim


def _variance_muxyhat(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray) -> np.ndarray:
    """
    Computes the unweighted variance-covariance matrix of the observed matching patterns

    Args:
        muxyhat: an (X, Y) matrix of matching patterns
        n: an X-vector of margins
        m: an Y-vector of margins

    Returns:
        an (XY, XY) matrix of the variance-covariance of the matching patterns
    """
    n_households = np.sum(n) + np.sum(m) - np.sum(muxyhat)
    mu_norm = (muxyhat / n_households).ravel()
    var_norm = np.diag(mu_norm) - np.outer(mu_norm, mu_norm)
    return n_households * var_norm


def _make_phi_mat(phi_bases: np.ndarray) -> np.ndarray:
    """Reshapes the array of bases to a matrix.

    Args:
        phi_bases: an (X, Y, K) array of bases

    Returns:
        the same,  (XY, K)-reshaped
    """
    X, Y, K = phi_bases.shape
    n_cells = X * Y
    phi_mat = np.zeros((n_cells, K))
    for k in range(K):
        phi_mat[:, k] = phi_bases[:, :, k].ravel()
    return phi_mat


def _check_dimensions(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray,
                      phi_bases: np.ndarray) -> Tuple[int, int, int]:
    """Check that the dimensions are consistent and return them.

    Args:
       muxyhat: should be a (X,Y) matrix
       n: should be a X-vector
       m: should be a Y-vector
       phi_bases: should be a (X,Y,K) array

    Returns:
       (X,Y,K) the shape of phi_bases
    """
    if phi_bases.ndim != 3:
        bs_error_abort(
            "phi_bases should have 3 dimensions not {phi_bases.ndim}")
    X, Y, K = phi_bases.shape
    Xm, Ym = test_matrix(muxyhat)
    if Xm != X or Ym != Y:
        bs_error_abort(
            "phi_bases is {phi_bases.shape} and muxyhat is {muxyhat.shape}.")
    Xn = test_vector(n)
    Yn = test_vector(m)
    if Xn != X:
        bs_error_abort(
            "muxyhat is {muxyhat.shape}  and n has {Xn} elements.")
    if Yn != Y:
        bs_error_abort(
            "muxyhat is {muxyhat.shape}  and m has {Yn} elements.")
    return X, Y, K


def _reshape4_to2(array4: np.ndarray) -> np.ndarray:
    """Reshapes an array (X,Y,Z,T) to a matrix (XY,ZT).

    Args:
        array4: an (X, Y, Z, T) array

    Returns:
        the same,  (XY, ZT)-reshaped
    """
    if array4.ndim != 4:
        bs_error_abort(
            "array4 should have 4 dimensions not {array4.ndim}")
    X, Y, Z, T = array4.shape
    XY, ZT = X * Y, Z*T
    array2 = np.zeros((XY, ZT))
    xy = 0
    for x in range(X):
        for y in range(Y):
            array2[xy, :] = array4[x, y, :, :].ravel()
            xy += 1
    return array2


def _change_indices(nests: List[List[int]]):
    """subtracts 1 from the indices within the nest structure

    Args:
        nests: the nest structure

    Returns:
        a similar list
    """
    return [
        [nest_i - 1 for nest_i in nest]
        for nest in nests
    ]


def _find_nest_of(nests: List[List[int]], y: int) -> int:
    """ find the index of the nest that contains y, or return -1

    Args:
        nests: a nest structure
        y: the type we are looking for

    Returns:
        the nest of y, or -1 if not found
    """
    for i_n, nest in enumerate(nests):
        if y in nest:
            return i_n
    return -1           # if not found
