""" Estimates semilinear separable models with a given entropy function.
The entropy function and the surplus matrix must both be linear in the parameters.
"""
import numpy as np
import scipy.linalg as spla
import scipy.stats as sts
from typing import Optional, List
from numpy.random import default_rng

from .utils import test_vector, test_matrix, bs_error_abort, print_stars
from .matching_utils import _make_phi_mat, _variance_muxyhat
from .model_classes import ChooSiowPrimitives
from .entropy import EntropyFunctions, _numeric_hessian, _fill_hessian_from_components
from .choo_siow import entropy_choo_siow, entropy_choo_siow_numeric
from .choo_siow_heteroskedastic import entropy_choo_siow_heteroskedastic, \
    entropy_choo_siow_heteroskedastic_numeric
from .choo_siow_gender_heteroskedastic import entropy_choo_siow_gender_heteroskedastic, \
    entropy_choo_siow_gender_heteroskedastic_numeric
from .min_distance_utils import MDEResults, _compute_estimates
from .nested_logit import _make_entropies_nested_logit


def estimate_semilinear_mde(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray,
                            phi_bases: np.ndarray,
                            entropy: EntropyFunctions,
                            more_params: Optional[List] = None,
                            initial_weights: Optional[np.ndarray] = None) -> MDEResults:
    """
    Estimates the parameters of the distributions and of the base functions.

    Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        entropy: an ``EntropyFunctions`` object
        more_params: additional parameters of the distribution of errors, if any
        initial_weights: if specified, used as the weighting matrix for the first step when `param_entropy` is `True`

    Returns:
        an `MDEResults` instance

    Example:
        see `mde_example` below
    """
    X, Y = test_matrix(muxyhat)
    Xn = test_vector(n)
    if Xn != X:
        bs_error_abort(f"n has {Xn} elements while muxyhat is ({X}, {Y})")
    Ym = test_vector(m)
    if Ym != Y:
        bs_error_abort(f"m has {Ym} elements while muxyhat is ({X}, {Y})")
    ndims_phi = phi_bases.ndim
    if ndims_phi != 3:
        bs_error_abort(f"phi_bases should have 3 dimensions, not {ndims_phi}")
    Xp, Yp, K = phi_bases.shape
    if Xp != X or Yp != Y:
        bs_error_abort(
            f"phi_bases should have shape ({X}, {Y}, {K}) not ({Xp}, {Yp}, {K})")
    parameterized_entropy = entropy.parameter_dependent
    if parameterized_entropy:
        if initial_weights is None:
            print_stars(
                "Using the identity matrix as weighting matrix in the first step.")
            S_mat = np.eye(X * Y)
        else:
            S_mat = initial_weights

    phi_mat = _make_phi_mat(phi_bases)
    e0_fun = entropy.e0_fun
    if more_params is None:
        e0_vals = e0_fun(muxyhat, n, m)
    else:
        e0_vals = e0_fun(muxyhat, n, m, more_params)
    e0_hat = e0_vals.ravel()

    if not parameterized_entropy:
        n_pars = K
        hessian = entropy.hessian
        if hessian == "provided":
            e0_derivative = entropy.e0_derivative
            if more_params is None:
                der2_entropy_hat = e0_derivative(muxyhat, n, m)
            else:
                der2_entropy_hat = e0_derivative(muxyhat, n, m, more_params)
        else:
            if more_params is None:
                der2_entropy_hat = _numeric_hessian(entropy, muxyhat, n, m)
            else:
                der2_entropy_hat = _numeric_hessian(
                    entropy, muxyhat, n, m, more_params)
        der2_entropy_hat = _fill_hessian_from_components(der2_entropy_hat)

        var_muxy = _variance_muxyhat(muxyhat, n, m)
        var_entropy_gradient = der2_entropy_hat.T @ var_muxy @ der2_entropy_hat
        # print(spla.eigvals(var_entropy_gradient))
        S_mat = spla.inv(var_entropy_gradient)
        estimated_coefficients, varcov_coefficients = _compute_estimates(
            phi_mat, S_mat, e0_hat)
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ estimated_coefficients
        residuals = est_Phi + e0_hat
    else:  # parameterized entropy
        # create the F matrix
        e_fun = entropy.e_fun
        if more_params is None:
            e_vals = e_fun(muxyhat, n, m)
        else:
            e_vals = e_fun(muxyhat, n, m, more_params)
        e_hat = _make_phi_mat(e_vals)

        F_hat = np.column_stack((e_hat, phi_mat))
        n_pars = e_hat.shape[1] + K
        # first pass with an initial weighting matrix
        first_coeffs, _ = _compute_estimates(F_hat, S_mat, e0_hat)
        first_alpha = first_coeffs[:-K]

        # compute the efficient weighting matrix
        hessian = entropy.hessian
        if hessian == "provided":
            e0_derivative = entropy.e0_derivative
            e_derivative = entropy.e_derivative
            if more_params is None:
                der2_e0_hat = e0_derivative(muxyhat, n, m)
                der2_e_hat = e_derivative(muxyhat, n, m)
            else:
                der2_e0_hat = e0_derivative(muxyhat, n, m, more_params)
                der2_e_hat = e_derivative(muxyhat, n, m, more_params)

            # print_stars("First-stage estimates:")
            # print(first_coeffs)

            first_alpha = first_coeffs[:-K]

            der2_entropy_hat = []
            for c in [0, 1, 2]:
                der2_entropy_hat.append(
                    der2_e0_hat[c] + der2_e_hat[c] @ first_alpha)
            der2_entropy_hat = tuple(der2_entropy_hat)
        else:
            if more_params is None:
                der2_entropy_hat = _numeric_hessian(
                    entropy, muxyhat, n, m, alpha=first_alpha)
            else:
                der2_entropy_hat = _numeric_hessian(entropy, muxyhat, n, m, alpha=first_alpha,
                                                    more_params=more_params)
        der2_entropy_hat = _fill_hessian_from_components(der2_entropy_hat)

        var_muxy = _variance_muxyhat(muxyhat, n, m)
        var_entropy_gradient = der2_entropy_hat.T @ var_muxy @ der2_entropy_hat
        S_mat = spla.inv(var_entropy_gradient)

        # second pass
        estimated_coefficients, varcov_coefficients = _compute_estimates(
            F_hat, S_mat, e0_hat)
        est_alpha, est_beta = estimated_coefficients[:-
                                                     K], estimated_coefficients[-K:]
        stderrs_coefficients = np.sqrt(np.diag(varcov_coefficients))
        est_Phi = phi_mat @ est_beta
        residuals = est_Phi + e0_hat + e_hat @ est_alpha

    value_obj = residuals.T @ S_mat @ residuals
    ndf = X * Y - n_pars
    test_stat = value_obj
    n_households = np.sum(n) + np.sum(m) - np.sum(muxyhat)

    results = MDEResults(X=X, Y=Y, K=K, number_households=n_households,
                         estimated_coefficients=estimated_coefficients,
                         varcov_coefficients=varcov_coefficients,
                         stderrs_coefficients=stderrs_coefficients,
                         estimated_Phi=est_Phi.reshape((X, Y)),
                         test_statistic=test_stat,
                         ndf=ndf,
                         test_pvalue=sts.chi2.sf(test_stat, ndf),
                         parameterized_entropy=parameterized_entropy)
    return results


def mde_example():

    X, Y, K = 5, 6, 2
    n_households = 1e4
    # we simulate a Choo and Siow population with equal numbers of men and women of each type
    lambda_true = np.random.randn(K)
    phi_bases = np.random.randn(X, Y, K)
    n = np.ones(X)
    m = np.ones(Y)
    Phi = phi_bases @ lambda_true
    choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
    mus_sim = choo_siow_instance.simulate(n_households)
    muxy_sim, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()

    nests_for_each_x = [[0, 1, 2], [3, 4]]
    nests_for_each_y = [[0, 1, 2], [3, 4, 5]]
    entropy_nested_logit, entropy_nested_logit_numeric = _make_entropies_nested_logit(
        nests_for_each_x, nests_for_each_y)

    entropy_models = [entropy_choo_siow, entropy_choo_siow_numeric,
                      entropy_choo_siow_gender_heteroskedastic,
                      entropy_choo_siow_gender_heteroskedastic_numeric,
                      entropy_choo_siow_heteroskedastic,
                      entropy_choo_siow_heteroskedastic_numeric,
                      entropy_nested_logit,
                      entropy_nested_logit_numeric]

    nests_params = [nests_for_each_x, nests_for_each_y]
    additional_params = [None]*6 + [nests_params]*2

    for (entropy_model, more_params) in zip(entropy_models, additional_params):
        if entropy_model.description is not None:
            print_stars(entropy_model.description)

        mde_results = estimate_semilinear_mde(muxy_sim, n_sim, m_sim, phi_bases,
                                              entropy_model, more_params=more_params)
        estimates = mde_results.estimated_coefficients
        stderrs = mde_results.stderrs_coefficients

        if entropy_model == entropy_choo_siow_gender_heteroskedastic:
            true_alpha = np.ones(1)

        if entropy_model == entropy_choo_siow_heteroskedastic:
            true_alpha = np.ones(estimates.size - lambda_true.size)

        if entropy_model in [entropy_nested_logit, entropy_nested_logit_numeric]:
            true_alpha = np.ones(estimates.size - lambda_true.size)

        if entropy_model.parameter_dependent:
            true_coeffs = np.concatenate((true_alpha, lambda_true))
        else:
            true_coeffs = lambda_true

        repr_str = "The true and estimated coefficients (and their standard errors) are\n\n"
        n_alpha = 0
        if entropy_model.parameter_dependent:
            n_alpha = true_alpha.size
            for i, coeff in enumerate(estimates[:n_alpha]):
                repr_str += f"   alpha({i + 1}): {true_alpha[i]: > 10.3f}    {coeff: > 10.3f}  ({stderrs[i]: > 10.3f})\n"
            repr_str += "\n"
        for i, coeff in enumerate(estimates[n_alpha:]):
            j = n_alpha + i
            repr_str += f"   base {i + 1}: {true_coeffs[j]: > 10.3f}    {coeff: > 10.3f}  ({stderrs[j]: > 10.3f})\n"
        repr_str += "\nSpecification test:\n"
        repr_str += f"   the value of the test statistic is {mde_results.test_statistic: > 10.3f}\n"
        repr_str += f"     for a chi2({mde_results.ndf}), the p-value is {mde_results.test_pvalue: > 10.3f}\n"
        print_stars(repr_str)
