from dataclasses import dataclass
import numpy as np
from typing import Optional, List

from .utils import test_vector, test_matrix, bs_error_abort, print_stars
from .matching_utils import Matching, _simulate_sample_from_mus, _find_nest_of
from .ipfp_solvers import ipfp_homo_solver, ipfp_nested_logit_solver


@dataclass
class ChooSiowPrimitives:
    Phi: np.ndarray
    n: np.ndarray
    m: np.ndarray
    mus: Optional[Matching] = None

    def __repr__(self):
        X, Y = self.Phi.shape
        nmen, nwomen = np.sum(self.n), np.sum(self.m)
        print_stars(f"This is a Choo and Siow market with {nmen} men of {X} types  and {nwomen} women of {Y} types.")

    def __post_init__(self):
        X, Y = test_matrix(self.Phi)
        Xn = test_vector(self.n)
        Ym = test_vector(self.m)
        if Xn != X:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but m has {Ym} elements.")

    def ipfp_solve(self) -> Matching:
        mus, err_x, err_y = ipfp_homo_solver(self.Phi, self.n, self.m)
        return Matching(muxy=mus[0], mux0=mus[1], mu0y=mus[2])

    def simulate(self, n_households: int) -> Matching:
        self.mus = self.ipfp_solve()
        mus_sim, _, _ = _simulate_sample_from_mus(self.mus, n_households)
        return mus_sim


@dataclass
class NestedLogitPrimitives:
    Phi: np.ndarray
    n: np.ndarray
    m: np.ndarray
    nests_for_each_x: List[List[int]]
    nests_for_each_y: List[List[int]]
    n_nests_over_X: int
    n_nests_over_Y: int
    nests_of_x: List[int]
    nests_of_y: List[int]
    n_alphas: int
    mus: Optional[Matching] = None
    true_alphas: Optional[np.ndarray] = None

    def __init__(self, Phi: np.ndarray, n: np.ndarray, m: np.ndarray,
                 nests_for_each_x: List[List[int]],
                 nests_for_each_y: List[List[int]],
                 true_alphas: Optional[np.ndarray]=None):
        """
        We only model two-level nested logit, with {0} as the first nest,
        and nests and nests parameters that do not depend on the type.

        Args:
            Phi: the (X,Y) joint surplus matrix
            n: the X-vector of men margins
            m: the X-vector of women margins
            nests_for_each_x: the composition of the nests over 1...Y, a list of r lists
            nests_for_each_y: the composition of the nests over 1...X, a list of s lists
            true_alphas: the true nest parameters, if any; should be an (r+s)-vector
        """
        X, Y = test_matrix(Phi)
        Xn = test_vector(n)
        Ym = test_vector(m)

        self.n_nests_over_X = len(nests_for_each_y)
        self.n_nests_over_Y = len(nests_for_each_x)
        self.n_alphas = self.n_nests_over_X+self.n_nests_over_Y

        if Xn != X:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but n has {Xn} elements.")
        if Ym != Y:
            bs_error_abort(f"Phi is a ({X}, {Y}) matrix but m has {Ym} elements.")

        if true_alphas is not None:
            alpha_size = test_vector(true_alphas)
            if alpha_size != self.n_alphas:
                bs_error_abort(f"true_alphas shoud have {self.n_alphas} elements, not {alpha_size}")

        self.Phi = Phi
        self.n = n
        self.m = m
        self.true_alphas = true_alphas
        self.nests_for_each_x = nests_for_each_x
        self.nests_for_each_y = nests_for_each_y

        # check that every x is in a nest, and just once
        nests_of_x = []
        for x in range(X):
            nests_of_x.append(_find_nest_of(nests_for_each_y, x))
        if -1 in nests_of_x or len(set(nests_of_x)) != len(nests_for_each_y):
            bs_error_abort("Check your nests_for_each_y")
        # check that every y is in a nest, and just once
        nests_of_y = []
        for y in range(Y):
            nests_of_y.append(_find_nest_of(nests_for_each_x, y))
        if -1 in nests_of_y or len(set(nests_of_y)) != len(nests_for_each_x):
            bs_error_abort("Check your nests_for_each_x")

        self.nests_of_x = nests_of_x
        self.nests_of_y = nests_of_y

    def __repr__(self):
        X, Y = self.Phi.shape
        nmen, nwomen = np.sum(self.n), np.sum(self.m)
        repr_str = f"This is a 2-level nested logit with {nmen} men of {X} types  and {nwomen} women of {Y} types.\n"
        repr_str += f" We have {self.n_nests_over_Y} nests over 1...Y and {self.n_nests_over_X} nests over 1...X,\n"
        if self.true_alphas is None:
            repr_str += "     with unspecified nests parameters."
        else:
            alpha_vals = self.true_alphas
            repr_str += "     with respective nests parameters:\n"
            repr_str += f"    {alpha_vals[:self.n_nests_over_Y]}\n"
            repr_str += f" and {alpha_vals[self.n_nests_over_Y:]}\n"
        print_stars(repr_str)


    def ipfp_solve(self) -> Matching:
        if self.true_alphas is None:
            bs_error_abort("true_alphas must be specified to solve the nested logit by IPFP.")
        mus, err_x, err_y = ipfp_nested_logit_solver(self.Phi, self.n, self.m,
                                                     self.nests_for_each_x,
                                                     self.nests_for_each_y,
                                                     self.true_alphas)
        return mus

    def simulate(self, n_households: int) -> Matching:
        self.mus = self.ipfp_solve()
        mus_sim, _, _ = _simulate_sample_from_mus(self.mus, n_households)
        return mus_sim

