"""Estimates the semilinear Choo and Siow homoskedastic (2006) model using Poisson GLM.
"""


import numpy as np
import scipy.linalg as spla
import scipy.sparse as spr
from sklearn import linear_model
from typing import Optional
from dataclasses import dataclass

from .choo_siow import _entropy_choo_siow
from .model_classes import ChooSiowPrimitives
from .utils import print_stars
from .matching_utils import _get_singles, _make_phi_mat, _check_dimensions, _variance_muxyhat, \
    _reshape4_to2
from .ipfp_solvers import ipfp_homo_solver


@dataclass
class PoissonGLMResults:
    """Stores and formats the estimation results.    """
    X: int
    Y: int
    K: int
    number_households: int
    number_individuals: int
    estimated_coefficients: np.ndarray
    estimated_Phi: np.ndarray
    estimated_u: np.ndarray
    estimated_v: np.ndarray
    variance_coefficients: np.ndarray
    stderrs_coefficients: np.ndarray

    def __repr__(self):
        line_stars = '*' * 80 + "\n"
        print_stars("Estimating a Choo and Siow model by Poisson GLM.")
        model_str = f"The data has {self.number_households} households\n\n"
        model_str += f"We use {self.K} basis functions.\n\n"
        repr_str = line_stars + model_str
        repr_str += "The estimated coefficients (and their standard errors) are\n\n"
        for i, coeff in enumerate(self.estimated_coefficients):
            repr_str += f"   base {i + 1}: {coeff: > 10.3f}  ({self.stderrs_coefficients[i]: .3f})\n"
        repr_str += "The estimated utilities of men are\n\n"
        for i, u_val in enumerate(self.estimated_u):
            repr_str += f"   base {i + 1}: {u_val: > 10.3f}\n"
        repr_str += "The estimated utilities of men are\n\n"
        for i, v_val in enumerate(self.estimated_v):
            repr_str += f"   base {i + 1}: {v_val: > 10.3f}\n"
        return repr_str + line_stars


def _prepare_data(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray):
    """Normalizes the matching patterns and stacks them.

       Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        entropy: an ``EntropyFunctions`` object
        initial_omega: if specified, used as the weighting matrix for the first step when `param_entropy` is `True`

       Returns:
        an `MDEResults` instance

       Example:
        See the `poisson_examples` function in the source of the module.
    """
    muxy = muxyhat.copy().astype(float)
    mux0hat, mu0yhat = _get_singles(muxy, n, m)
    n_households = muxy.sum() + mux0hat.sum() + mu0yhat.sum()
    n_individuals = n_households + muxy.sum()
    # rescale the data so that the total number of individuals is one
    muhat_norm = np.concatenate(
        [muxyhat.flatten(), mux0hat, mu0yhat]) / n_individuals
    return muhat_norm, n_households, n_individuals


def choo_siow_poisson_glm(muxyhat: np.ndarray, n: np.ndarray, m: np.ndarray, phi_bases: np.ndarray,
                          tol: Optional[float] = 1e-12, max_iter: Optional[int] = 10000):
    """
    Estimates the semilinear Choo and Siow homoskedastic (2006) model using Poisson GLM.

       Args:
        muxyhat: an (X, Y) matrix of observed matching patterns
        n: an X-vector of observed margins
        m: an Y-vector of observed margins
        phi_bases: an (X, Y, K) array of bases
        tol: tolerance level for ``linear_model.PoissonRegressor.fit``
        max_iter: maximum number of iterations for ``linear_model.PoissonRegressor.fit``
       Returns:
        a `PoissonGLMResults` instance

       Example:
        See the `main` function in the source of the module.
    """
    X, Y, K = _check_dimensions(muxyhat, n, m, phi_bases)

    XY = X * Y
    # the vector of weights for the Poisson regression
    d_a = np.append(2 * np.ones(XY), np.ones(X + Y))
    # the (XY+X+Y, X+Y) matrix
    B = spr.vstack([
        spr.hstack([
            spr.kron(spr.identity(X), np.ones((Y, 1))), spr.kron(
                np.ones((X, 1)), spr.identity(Y))
        ]),
        spr.hstack([
            spr.identity(X), spr.csr_matrix((X, Y))
        ]),
        spr.hstack([
            spr.csr_matrix((Y, X)), spr.identity(Y)
        ])
    ])
    # reshape the bases into an (XY+X+Y, K) matrix
    phi_mat = _make_phi_mat(phi_bases)
    phi_stacked = np.vstack(
        (phi_mat, np.zeros((X, K)), np.zeros((Y, K))))
    # finally, the (XY+X+Y, X+Y+K) matrix C
    C = spr.hstack([-B, phi_stacked])

    muhat_norm, n_households, n_individuals = _prepare_data(muxyhat, n, m)

    clf = linear_model.PoissonRegressor(
        fit_intercept=False, tol=tol, verbose=3, alpha=0, max_iter=max_iter)
    clf.fit(spr.diags(1 / d_a) @ C, muhat_norm, sample_weight=d_a)

    beta_est = clf.coef_[-K:]

    # we compoute the variance-covariance of the estimator
    varmu = _variance_muxyhat(muxyhat, n, m)
    _, _, hessian_E = _entropy_choo_siow(muxyhat, n, m, deriv=2)
    hessian_mat = _reshape4_to2(hessian_E)
    inverse_E = spla.inv(hessian_mat)
    big_mat = spla.inv(phi_mat.T @ inverse_E @ phi_mat)
    variance_coefficients = big_mat.T @ (phi_mat.T @ varmu @ phi_mat) @ big_mat
    stderrs_coefficients = np.sqrt(np.diag(variance_coefficients))

    Phi_est = phi_bases @ beta_est

    # we correct for the effect of the normalization
    u_est = clf.coef_[:X]  + np.log(n/n_individuals)
    v_est = clf.coef_[X: -K] + np.log(m/n_individuals)

    results = PoissonGLMResults(X=X, Y=Y, K=K, number_households=n_households, number_individuals=n_individuals,
                                estimated_coefficients=beta_est, estimated_Phi=Phi_est,
                                estimated_u=u_est, estimated_v=v_est,
                                variance_coefficients=variance_coefficients,
                                stderrs_coefficients=stderrs_coefficients)

    return results


def poisson_examples():

    # Example 1: using IPFP to solve for the matching

    X, Y, K = 4, 3, 6
    phi_bases = np.zeros((X, Y, K))
    phi_bases[:, :, 0] = 1
    for x in range(X):
        phi_bases[x, :, 1] = x
        phi_bases[x, :, 3] = x*x
        for y in range(Y):
            phi_bases[x, y, 4] = x*y
    for y in range(Y):
        phi_bases[:, y, 2] = y
        phi_bases[:, y, 5] = y*y

    lambda_true = np.array([2.31, 0.0, 0.41, 0.0, 0.19, 0.0])
    Phi = phi_bases @ lambda_true

    nx = np.array([5, 6, 5, 8])*1000
    my = np.array([4, 8, 6])*1000
    (muxy, mux0, mu0y), err_x, err_y = ipfp_homo_solver(Phi, nx, my)
    results = choo_siow_poisson_glm(muxy, nx, my, phi_bases)
    print(results)

    print_stars("The true and estimated coefficients are:")
    print(np.column_stack((lambda_true, results.estimated_coefficients)))

    print_stars("The true and estimated utilities for men are:")
    print(np.column_stack((-np.log(mux0/nx), results.estimated_u)))

    print_stars("The true and estimated utilities for women are:")
    print(np.column_stack((-np.log(mu0y/my), results.estimated_v)))

    # Example 2: simulating many individuals

    X, Y, K = 5, 6, 2
    n_households = 1e4
    # we simulate a Choo and Siow population with equal numbers of men and women of each type
    lambda_true = np.random.randn(K)
    phi_bases = np.random.randn(X, Y, K)
    n = np.ones(X)
    m = np.ones(Y)
    Phi = phi_bases @ lambda_true
    choo_siow_instance = ChooSiowPrimitives(Phi, n, m)
    mus_sim  = choo_siow_instance.simulate(n_households)
    muxy_sim, mux0_sim, mu0y_sim, n_sim, m_sim = mus_sim.unpack()

    results = choo_siow_poisson_glm(muxy_sim, n_sim, m_sim, phi_bases)
    print(results)

    print_stars("The true and estimated coefficients are:")
    print(np.column_stack((lambda_true, results.estimated_coefficients)))

    print_stars("The true and estimated utilities for men are:")
    print(np.column_stack((-np.log(mux0_sim/n_sim), results.estimated_u)))

    print_stars("The true and estimated utilities for women are:")
    print(np.column_stack((-np.log(mu0y_sim/m_sim), results.estimated_v)))
